{ ----------------------- SIMULA ------------------------ }

UNIT Simula ;

INTERFACE
{ $N+, E+, F+ }

TYPE
	 RealType = real ;

CONST
	  ArraySize = ( 2 * MaxInt ) div SizeOf ( RealType ) ;

TYPE
	 ArrayType = array [ 1..ArraySize ] of RealType ;
	  ProcType = PROCEDURE ( T : RealType ; VAR Y, YP ) ;

PROCEDURE RKF45 (    F : ProcType ;
						NEQN : integer ;
					  VAR Y ;
					  VAR T : RealType ;
						Tout : RealType ;
				VAR RelErr : RealType ;
					 AbsErr : RealType ;
				 VAR Iflag : integer ;
					  VAR Work, IworkI ) ;

PROCEDURE Runge (    F : ProcType ;
						NEQN : integer ;
					  VAR YI ;
					  VAR T : RealType ;
						Tout : RealType ;
							H : RealType ;
					  VAR Work ) ;

IMPLEMENTATION

FUNCTION Amax1 ( x, y : RealType ) : RealType ;
BEGIN { Amax1 }
	  IF x > y then Amax1 := x else Amax1 := y ;
END { Amax1 } ;

PROCEDURE RKF45 (    F : ProcType ;
						NEQN : integer ;
					  VAR Y ;
					  VAR T : RealType ;
						Tout : RealType ;
				VAR RelErr : RealType ;
					 AbsErr : RealType ;
				 VAR Iflag : integer ;
					  VAR Work, IworkI ) ;

VAR
	W : ArrayType absolute Work ;
	IWork : array [ 1..5 ] of integer absolute IworkI ;
	k1, k2, k3, k4, k5, k6, k1m : integer ;

PROCEDURE RKFS (    F : ProcType ;
					  NEQN : integer ;
					 VAR YI ;
					 VAR T : RealType ;
					 Tout, RelErr, AbsErr : RealType ;
					 VAR Iflag : integer ;
					 VAR YPI, HI, F1I, F2I, F3I, F4I, F5I ;
					 VAR SaveREI, SaveAEI, NFEI, KOPI ;
					 VAR InitI, JflagI, KflagI ) ;

VAR
	Y  : ArrayType absolute YI ;
	YP : ArrayType absolute YPI ;
	H  : RealType absolute HI ;
	F1 : ArrayType absolute F1I ;
	F2 : ArrayType absolute F2I ;
	F3 : ArrayType absolute F3I ;
	F4 : ArrayType absolute F4I ;
	F5 : ArrayType absolute F5I ;
	SaveRE : RealType absolute SaveREI ;
	SaveAE : RealType absolute SaveAEI ;
	NFE    : integer absolute NFEI ;
	KOP    : integer absolute KOPI ;
	Init   : integer absolute InitI ;
	Jflag  : integer absolute JflagI ;
	Kflag  : integer absolute KflagI ;

CONST
	  REMin = 1.0E-12 ;
	  MaxNFE = 3000 ;

VAR
	A, AE, DT, EE, EEOET, EstTol, ET, Hmin, RER,
	S, Scale, Tol, TolN, U26, EPS, YPK : RealType ;
	k, Mflag : integer ;
	Hfaild, Output : boolean ;

LABEL
	  10, 40, 45, 50, 60, 65, 80, 100, 200 ;

PROCEDURE FEHL (    F : ProcType ;
					  NEQN : integer ;
					 VAR YI ;
					 VAR T : RealType ;
						  H : RealType ;
					 VAR YPI, F1I, F2I, F3I, F4I, F5I, SI ) ;

VAR
	Y  : ArrayType absolute YI ;
	YP : ArrayType absolute YPI ;
	F1 : ArrayType absolute F1I ;
	F2 : ArrayType absolute F2I ;
	F3 : ArrayType absolute F3I ;
	F4 : ArrayType absolute F4I ;
	F5 : ArrayType absolute F5I ;
	S  : ArrayType absolute SI ;

VAR
	CH : RealType ;
	k  : integer ;

BEGIN { FEHL }

	  CH := H / 4 ;
	  for k := 1 to NEQN do
		  F5[ k ] := Y[ k ] + CH * YP[ k ] ;
	  F( T + CH, F5, F1 ) ;
	  CH := 3. * H / 32. ;
	  for k := 1 to NEQN do
		  F5[ k ] := Y[ k ] + CH * ( YP[ k ] + 3. * F1[ k ] ) ;
	  F( T + 3. * H / 8., F5, F2 ) ;
	  CH := H / 2197. ;
	  for k := 1 to NEQN do
		  F5[ k ] := Y[ k ] + CH *
					 ( 1932. * YP[ k ] + ( 7296. * F2[ k ] - 7200. * F1[ k ] ) ) ;
	  F( T + 12. * H / 13., F5, F3 ) ;
	  CH := H / 4104. ;
     for k := 1 to NEQN do
		  F5[ k ] := Y[ k ] + CH *
					 ( ( 8341. * YP[ k ] - 845. * F3[ k ] ) +
					 ( 29440. * F2[ k ] - 32832. * F1[ k ] ) ) ;
	  F( T + H, F5, F4 ) ;
	  CH := H / 20520. ;
     for k := 1 to NEQN do
		  F1[ k ] := Y[ k ] + CH *
					 ( ( -6080. * YP[ k ] + ( 9295. * F3[ k ] - 5643. * F4[ k ] ) ) +
					 ( 41040. * F1[ k ] - 28352. * F2[ k ] ) ) ;
	  F( T + H / 2., F1, F5 ) ;
	  CH := H / 7618050. ;
     for k := 1 to NEQN do
		  S[ k ] := Y[ k ] + CH *
					 ( ( 902880. * YP[ k ] + ( 3855735. * F3[ k ] - 1371249. * F4[ k ] ) ) +
					 ( 3953664. * F2[ k ] + 277020. * F5[ k ] ) ) ;
END { FEHL } ;

BEGIN { RKFS }

	  Mflag := abs( Iflag ) ;
	  IF ( NEQN < 1 ) or ( RelErr < 0 ) or ( AbsErr < 0 )
		  or ( Mflag = 0 ) or ( Mflag > 8 ) then
			 BEGIN
10 :           Iflag := 8 ;
					exit
			 END ;
	  IF Mflag = 1 then
		 BEGIN
				EPS := 1. ;
				while 1. + EPS <> 1 do EPS := EPS / 2. ;
				U26 := 26. * EPS ;
		 END
	  ELSE
			BEGIN
				  IF ( T = Tout ) and ( Kflag <> 3 ) then goto 10 ;
				  IF Mflag = 2 then
					 BEGIN
							IF ( Kflag = 3 ) or ( Init = 0 ) then goto 45 ;
							IF Kflag = 4 then goto 40 ;
							IF ( Kflag = 5 ) and ( AbsErr = 0 ) or
								( Kflag = 6 ) and ( RelErr <= SaveRE )
								and ( AbsErr <= SaveAE ) then HALT ;
							goto 50 ;
					 END
				  ELSE
						BEGIN
							  IF ( Iflag = 3 ) or ( Iflag = 5 ) and ( AbsErr > 0 )
								 then goto 45 ;
							  IF Iflag = 4 then goto 40 ;
							  HALT
						END ;

40 :          NFE := 0 ;
				  IF Mflag = 2 then goto 50 ;

45 :          Iflag := Jflag ;
				  IF Kflag = 3 then Mflag := abs( Iflag )
			END ;

50 : Jflag := Iflag ;
	  Kflag := 0 ;

	  SaveRE := RelErr ;
	  SaveAE := AbsErr ;

	  RER := 2. * EPS + REMin ;
	  IF RelErr < RER then
		 BEGIN
				RelErr := RER ;
				Iflag := 3 ;
				Kflag := 3 ;
				EXIT
		 END ;
	  DT := Tout - T ;
	  IF Mflag = 1 then goto 60 ;
	  IF Init = 0 then goto 65 ;
	  goto 80 ;

60 : Init := 0 ;
	  KOP := 0 ;
	  A := T ;
	  F( A, Y, YP ) ;
	  NFE := 1 ;
	  IF T = Tout then
		 BEGIN
				Iflag := 2 ; EXIT
		 END ;

65 : Init := 1 ;
	  H := abs( DT ) ;
	  TolN := 0. ;
	  for k := 1 to NEQN do
		  BEGIN
				 Tol := RelErr * abs( Y[ k ] ) + AbsErr ;
				 IF Tol <> 0 then
					BEGIN
						  TolN := Tol ;
						  YPK := abs( Y[ k ] ) ;
						  IF YPK * H * H * H * H * H > Tol then
							 H := exp( 0.2 * ln( Tol / YPK ) )
					END
		  END ;
	  IF TolN <= 0 then H := 0. ;
	  H := Amax1( H, U26 * Amax1( abs( T ), abs( DT ) ) ) ;
	  IF Iflag < 0 then Jflag := -2 ELSE Jflag := 2 ;

80 : IF DT > 0 then H := abs( H ) ELSE H := - abs( H ) ;
	  IF abs( H ) >= 2. * abs( DT ) then inc( KOP ) ;
	  IF KOP >= 100 then
		 BEGIN
				KOP := 0 ;
				Iflag := 7 ;
				EXIT ;
		 END ;
	  IF abs( DT ) <= U26 * abs( T ) then
		 BEGIN
				for k := 1 to NEQN do Y[ k ] := Y[ k ] + DT * YP[ k ] ;
				A := Tout ;
				F( A, Y, YP ) ;
				inc( NFE ) ; T := Tout ; Iflag := 2 ;
				EXIT
		 END ;

	  Output := false ;
	  Scale := 2. / RelErr ;
	  AE := Scale * AbsErr ;

100 :  Hfaild := false ;
		 HMin := U26 * abs( T ) ;
		 DT := Tout - T ;
		 IF abs( DT ) < abs( H ) then
			BEGIN
				  H := DT ; Output := true
			END
		 ELSE
			  IF abs( DT ) < 2. * abs( H ) then H := DT / 2. ;

200 :  IF NFE > MaxNFE then
			BEGIN
				  Iflag := 4 ; Kflag := 4 ; EXIT
			END ;

		 FEHL ( F, NEQN, Y, T, H, YP, F1, F2, F3, F4, F5, F1 ) ;

		 NFE := NFE + 5 ;
		 EEOET := 0 ;
		 for k := 1 to NEQN do
			 BEGIN
					ET := abs( Y[ k ] ) + abs( F1[ k ] ) + AE ;
					IF ET <= 0. then
					  BEGIN
							 Iflag := 5 ; EXIT
					  END ;
					EE := abs( ( -2090. * YP[ k ] +
								  ( 21970. * F3[ k ] - 15048. * F4[ k ] ) ) +
								  ( 22528. * F2[ k ] - 27360. * F5[ k ] ) ) ;
					EEOET := Amax1( EEOET, EE / ET )
			 END ;
		 EstTol := abs( H ) * EEOET * Scale / 52440. ;
		 IF EstTol > 1. then
			BEGIN
				  Hfaild := true ;
				  Output := false ;
				  IF EstTol < 52049
					 then S := 0.9 / exp( 0.2 * ln( EstTol ) )
					 ELSE S := 0.1 ;
				  H := S * H ;
				  IF abs( H ) > HMin then goto 200
				  ELSE
						BEGIN
							  Iflag := 6 ; Kflag := 6 ; EXIT
						END
			END ;

		 T := T + H ;
		 for k := 1 to NEQN do Y[ k ] := F1[ k ] ;
		 A := T ;
		 F( A, Y, YP ) ;
		 inc( NFE ) ;
		 IF EstTol > 1.889568E-4
			then S := 0.9 / exp( 0.2 * ln( EstTol ) )
			ELSE S := 5. ;
		 IF Hfaild then if S > 1 then S := 1 ;
		 S := Amax1( S * abs( H ), HMin ) ;
		 IF H < 0 then H := -S ELSE H := S ;

		 IF not Output then
			IF Iflag > 0 then goto 100
			ELSE Iflag := -2
		 ELSE Iflag := 2
END { RKFS } ;

BEGIN { RKF45 }
	  k1m := NEQN + 1 ;
	  k1  := k1m + 1 ;
	  k2  := k1 + NEQN ;
	  k3  := k2 + NEQN ;
	  k4  := k3 + NEQN ;
	  k5  := k4 + NEQN ;
	  k6  := k5 + NEQN ;
	  RKFS ( F, NEQN, Y, T, Tout, RelErr, AbsErr, Iflag, w[ 1 ],
				w[ k1m ], w[ k1 ], w[ k2 ], w[ k3 ], w[ k4 ], w[ k5 ],
				w[ k6 ], w[ k6 + 1 ], IWork[ 1 ], IWork[ 2 ], IWork[ 3 ],
				IWork[ 4 ], IWork[ 5 ] )
END { RKF45 } ;

PROCEDURE Runge (    F : ProcType ;
						NEQN : integer ;
					  VAR YI ;
					  VAR T : RealType ;
						Tout : RealType ;
						H 	  : RealType ;
					  VAR Work ) ;

VAR
	W  : array[ 1..3, 1..ArraySize div 3 ] of RealType absolute Work ;
	Y  : ArrayType absolute YI ;
	k  : integer ;
	DT : RealType ;

CONST
	  EPS : RealType = 0 ;

BEGIN { Runge }

	  IF ( NEQN <= 0 ) or ( H = 0 ) then HALT ;
	  IF EPS = 0. then
		 BEGIN
				EPS := 1. ;
				while 1 + EPS / 2. <> 1. do EPS := EPS / 2. ;
				EPS := 20. * EPS
		 END ;

	  REPEAT

			  F( T + 0., Y, W ) ;
			  DT := Tout - T ;
			  IF abs( DT ) <= abs( T ) * EPS then
				 BEGIN
						for k := 1 to NEQN do Y[ k ] := Y[ k ] + DT * W[ 1, k ] ;
						T := Tout ;
						EXIT
				 END ;
			  IF abs( DT ) < abs( H ) then H := DT ;
			  IF DT < 0. then H := - abs( H ) ELSE H := abs( H ) ;
			  for k := 1 to NEQN do W[ 2, k ] := Y[ k ] + H * W[ 1, k ] / 2. ;
			  F( T + H / 2., W[ 2, 1 ], W[ 3, 1 ] ) ;
			  for k := 1 to NEQN do
				  BEGIN
						 W[ 1, k ] := W[ 1, k ] + 2. * W[ 3, k ] ;
						 W[ 2, k ] := Y[ k ] + H * W[ 3, k ] / 2. ;
				  END ;
			  F( T + H / 2., W[ 2, 1 ], W[ 3, 1 ] ) ;
           for k := 1 to NEQN do
				  BEGIN
						 W[ 1, k ] := W[ 1, k ] + 2. * W[ 3, k ] ;
						 W[ 2, k ] := Y[ k ] + H * W[ 3, k ] ;
				  END ;
			  F( T + H, W[ 2, 1 ], W[ 3, 1 ] ) ;
			  for k := 1 to NEQN do
				  Y[ k ] := Y[ k ] + H * ( W[ 1, k ] + W[ 3, k ] ) / 6.  ;
			  T := T + H ;

	  UNTIL T = Tout

END { Runge } ;

END.
{ -------------------- end of SIMULA --------------------- }
